# -*- coding: utf-8 -*-

import setuptools
import versioneer

with open("README.md", "r") as handle:
    long_description = handle.read()

if __name__ == "__main__":
    setuptools.setup(
        name='opt_einsum',
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        description='Optimizing numpys einsum function',
        author='Daniel Smith',
        author_email='dgasmith@icloud.com',
        url="https://github.com/dgasmith/opt_einsum",
        license='MIT',
        packages=setuptools.find_packages(),
        install_requires=[
            'numpy>=1.7',
        ],
        extras_require={
            'docs': [
                'sphinx==1.2.3',  # autodoc was broken in 1.3.1
                'sphinxcontrib-napoleon',
                'sphinx_rtd_theme',
                'numpydoc',
            ],
            'tests': [
                'pytest',
                'pytest-cov',
                'pytest-pep8',
            ],
        },

        tests_require=[
            'pytest',
            'pytest-cov',
            'pytest-pep8',
        ],

        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
        ],
        zip_safe=True,
        long_description=long_description,
        long_description_content_type="text/markdown"
    )
